class InvalidContextException(Exception):
        pass


class ValidatorNotCallableException(Exception):
        pass


class ValidationError(Exception):
        pass


class FormDataAccessException(Exception):
    def __str__(self):
        return "Attempted to validate data that was not passed in to the validation method. Check to ensure validation is happening on form submission"



import os

class JinjaRenderer(object):
    from jinja2 import Environment, FileSystemLoader

    # a defualt search path for templates
    search_path = [os.path.dirname(os.path.realpath(__file__)) +
            "/templates/jinja/"]

    # allow lazy loading of the enviroment
    _env = None

    # the default template suffix
    suffix = ".html"

    @property
    def env(self):
        """ Simple lazy loader for the Jinja2 enviroment """
        if not self._env:
            self._env = JinjaRenderer.Environment(
                    loader=JinjaRenderer.FileSystemLoader(JinjaRenderer.search_path))
        return self._env

    def render(self, nodes, g_context):
        """ Loop over each Node and render it into a big blob of a string """
        buildup = ""
        for node in nodes:
            template = self.env.get_template(node.template + self.suffix)
            buildup += template.render(node.get_context(g_context))
        return buildup



from yota.exceptions import InvalidContextException, FormDataAccessException


class Node(object):
    """ Nodes are holders of context for rendering and displaying validating for a portion of
    your :class:`Form`. This default base Node is designed to provide a template
    along with specific context information to a templating engine such as
    Jinja2. For validation a Node acts as an information source or an error
    sink. Essentially Nodes can be used to source data for use in a
    :class:`Check`, and they can then be delivered some sort of validation error
    via a the internal :attr:`errors` attribute.

    .. note:: By default all keyword attributes passed to a Node's init function are passed onto the rendering context. To override this, use the :attr:`Node._ignores` attribute.

    :param _attr_name: This is how the Node is identified in the Form. If populated automatically if the Node is defined in an a Form class definition, however if the Node is added dynamically it will need to be defined before adding it to the Form.
    :type _attr_name: string

    :param _ignores: A List of attribute names to explicity not include in the rendering context. Mostly a niceity for keeping the rendering context clutter free.
    :type _ignores: list

    :param _requires: A List of attributes that will be required at render time. An exception will be thrown if these attributes are not present. Useful for things like lists that require certain data to render properly.
    :type _requires: list

    :param template: String name of the template to be parsed upon rendering. This is passed into the `Form._renderer` so it needs to be whatever that is designed to accept. Jinja2 is looking for a filename like 'node' that occurs in it's search path.
    :type template: string

    :param validator: An optional attribute that specifies a :class:`Check` object to be associated with the Node. This is automatically extracted at parse time and cannot be manipulated after Node insertion.
    :type validator: callable

    The default Node init method accepts any keyword arguments and adds them to
    the Node's rendering context.

    """

    _create_counter = 0
    """ Allows tracking the order of Node creation """

    def __init__(self,
                 template=None,
                 validator=None,
                 label=True,
                 _requires=None,
                 _ignores=None,
                 _attr_name=None,
                 **kwargs):

        # Allow _ignores and _requires to be overwritten at the instance level,
        # and also have a default
        if _ignores:
            self.ignores = _ignores
        elif not hasattr(self, '_ignores'):  # careful not to overwrite class attr
            self._ignores = ['template', 'validator']

        if _requires:
            self._requires = _requires
        elif not hasattr(self, '_requires'):  # careful not to overwrite class attr
            self._requires = []  # make sure it's there to prevent errors

        if not hasattr(self, 'validator'):  # careful not to overwrite class attr
            self.validator = validator


        if template:
            self.template = template
        self._attr_name = _attr_name
        self.label = label

        # Allows the parent form to keep track of attribute order
        self._create_counter = Node._create_counter
        Node._create_counter += 1

        # A placeholder for validation process
        self.errors = []
        self.data = ''

        # passes everything to our rendering context and updates params
        self.__dict__.update(kwargs)

    def add_error(self, error):
        """ This method serves mostly as a wrapper alowing for different error
        ordering semantics, or possibly error post-processing. Errors from
        validation methods should be added in this way allowing them to be
        caught. More information about what gets passed in in the
        :doc:`Validators` section. """
        self.errors.append(error)

    def set_identifiers(self, parent_name):
        """ This function gets called by the parent `Form` when it is
        initialized or inserted. It is designed to set various unique
        identifiers. By default it generates an id for the Node that is
        {parent_name}_{_attr_id}, a title for the Node that is the _attr_name
        capitalized, and a name for the element that is just the _attr_name. All
        of these attributes are then passed onto the rendering context of the
        Node by default. By default all of these attributes will yield to
        attributes passed into the __init__ method.

        :param parent_name: The name of the parent form. Useful in ensuring unique identifiers on your element names.
        :type parent_name: string
        """

        # Set some good defaults based on attribute name and parent name,
        # but always allow the user to override the values at the init level
        if not hasattr(self, 'id'):
            self.id = "{0}_{1}".format(parent_name, self._attr_name)
        if not hasattr(self, 'name'):
            self.name = self._attr_name
        if not hasattr(self, 'title'):
            self.title = self._attr_name.capitalize().replace('_', ' ')

    def resolve_data(self, data):
        """ This method is called when resolving the data from a form submission
        and linking it to a specific Node. The return value of this function is
        passed directly to the Validators data portion for your node. By default
        this will try and lookup data from the submission using the name
        attribute. """
        try:
            self.data = data[self.name]
        except:
            raise FormDataAccessException

    def get_context(self, g_context):
        """ Builds our rendering context for the Node at render time. By default
        all attributes of the Node are added to the global namespace and the
        global rendering context is passed in under the variable 'g'. This
        function is designed to be overridden for customization.
        :param g_context: The global rendering context passed in from the rendering method.
        """

        # d = {i: getattr(self, i) for i in dir(self) if not i.startswith("_") and i not in self._ignores }
        # Dat 2.6 compat
        d = {}
        for key in dir(self):
            if not key.startswith("_") and key not in self._ignores:
                d[key] = getattr(self, key)

        # check to make sure all required attributes are present
        for r in self._requires:
            if r not in d:
                raise InvalidContextException(
                    "Missing required context value '{0}'".format(r))
        d['g'] = g_context
        return d


class BaseNode(Node):
    """ This base Node supplies the name of the base rendering template the
    is used for standard form elements. This base template provides error
    divs and the horizontal form layout for Bootstrap.
    """
    base = "horiz.html"
    css_class = ''
    css_style = ''


class ListNode(BaseNode):
    """ Node for providing a basic drop down list. Requires an attribute that
     is a list of tuples providing the key and value for the dropdown list
     items.
    """
    template = 'list'
    _requires = ['items']

class ButtonNode(BaseNode):
    """ Creates a simple button in your form.
    """
    template = 'button'

class EntryNode(BaseNode):
    template = 'entry'

class TextareaNode(BaseNode):
    template = 'textarea'
    rows = '5'
    columns = '10'

class SubmitNode(BaseNode):
    template = 'submit'

class LeaderNode(Node):
    """ A Node that simply removes the title attribute from the Node rendering
    context. Intended for use in the start and end Nodes. """

    def set_identifiers(self, parent_name):
        # set our start node's id to actually be the name of the form
        if not hasattr(self, 'id'):
            self.id = parent_name

        super(LeaderNode, self).set_identifiers(parent_name)
        delattr(self, 'title')



class PostProcessor(object):
    """ A base class for all post processors. Post
    processors handle interoperability between different
    web development frameworks. """
    pass


class FlaskPostProcessor(object):
    def filter_post(self, postdict):
        return postdict



from collections import namedtuple
import re
from yota.exceptions import FormDataAccessException


class MinLengthValidator(object):
    """ Checks to see if data is at least :length long.

    :param length: The minimum length of the data.
    :type length: integer

    :param message: The message to present to the user upon failure.
    :type message: string
    """
    __slots__ = ["message", "min_length"]

    def __init__(self, length, message=None):
        self.min_length = length
        self.message = message if message else "Minimum allowed length {0}" \
            .format(length)
        super(MinLengthValidator, self).__init__()

    def __call__(self, target):
        if len(target.data) < self.min_length:
            target.add_error({'message': self.message})


class MaxLengthValidator(object):
    """ Checks to see if data is at most :length long.

    :param length: The maximum length of the data.
    :type length: integer

    :param message: The message to present to the user upon failure.
    :type message: string
    """
    __slots__ = ["message", "max_length"]

    def __init__(self, length, message=None):
        self.max_length = length
        self.message = message if message else "Maximum allowed length {0}" \
            .format(length)
        super(MaxLengthValidator, self).__init__()

    def __call__(self, target):
        if len(target.data) > self.max_length:
            target.add_error({'message': self.message})


class NonBlockingDummyValidator(object):
    """ A dummy class for testing non-blocking validators
    """

    def __call__(self, target):
        target.add_error({'message': "I'm not blocking!", 'block': False})


class RequiredValidator(object):
    """ Checks to make sure the user entered something.

    :param message: (optional) The message to present to the user upon failure.
    :type message: string
    """
    __slots__ = ["message"]

    def __init__(self, message=None):
        self.message = message if message else "A value is required"
        super(RequiredValidator, self).__init__()

    def __call__(self, target=None):
        if len(target.data) == 0:
            target.add_error({'message': self.message})


class EmailValidator(object):
    """ A direct port of the Django Email validator. Checks to see if an
    email is valid using regular expressions.
    """

    user_regex = re.compile(
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*$"  # dot-atom
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"$)',
        # quoted-string
        re.IGNORECASE)
    domain_regex = re.compile(
        r'(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?$)'  # domain
        # literal form, ipv4 address (SMTP 4.1.3)
        r'|^\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$',
        re.IGNORECASE)
    domain_whitelist = ['localhost']

    def __init__(self, message=None):
        self.message = message if message else "Entered value must be a valid" \
                                               " Email address"
        super(EmailValidator, self).__init__()

    def valid(self, value):
        """ A small breakout function to make passing back errors less
        redundant.
        """
        if not value or '@' not in value:
            return False

        user_part, domain_part = value.rsplit('@', 1)
        if not self.user_regex.match(user_part):
            return False

        if (not domain_part in self.domain_whitelist and
                not self.domain_regex.match(domain_part)):
            # Try for possible IDN domain-part
            try:
                domain_part = domain_part.encode('idna').decode('ascii')
                if not self.domain_regex.match(domain_part):
                    return False
                else:
                    return True
            except UnicodeError:
                return False

        return True

    def __call__(self, target):
        if self.valid(target.data):
            return None
        else:
            target.add_error({'message': self.message})


class Check(object):
    """ This object wraps a validator callable and is intended to be used in
    your `Form` subclass definition.

    :param callable validator: This is required to be a callable object that will carry out the actual validation. Many generic validators exist or you can roll your own.

    :param list args: A list of strings, or a single string, representing that _attr_name of the `Node` you would like passed into the validator. Once a validator is called this string will get resolved into a NamedTuple that will be passed into the validator callable.

    :param dict kwargs: Same as args above except it allows passing in node information as keyword arguments to the validator callable.

    `Check` objects are designed to be declared in your form subclass in one of
    two ways. Explicit.

    .. code-block:: python

        class MyForm(yota.Form):
            # This syntax shortens up the above explicit syntax for simple
            # validators
            first = EntryNode(title='First name')
            _first_valid = Check(MinLengthValidator(5), 'first')

    The syntax above binds our `MinLengthValidator` to a Node with attribute
    name 'first'. How it matches up your string 'first' to the correct `Node`
    instance is a bit of magic behind the scenes, but just know that it matches
    up to the attribute name you give the node in the your `Form` class
    definition. If you later add a `Node` dynamically it will need to specify
    the _attr_name upon declaration explicitly. More on this in :ref:`Dynamically
    Adding Nodes`.

    Implicit declaration is a nice option for simple validators, and is simply
    syntactic sugar for the above syntax.

    .. code-block:: python

        class MyForm(yota.Form):
            # This syntax shortens up the above explicit syntax for simple
            # validators. An args of 'first' will automatically be added to the
            # Check object for you.
            first = EntryNode(title='First name',
                              validator=Check(MinLengthValidator(5)))

            # This even more brief syntax will automatically built the Check
            # object for you
            last = EntryNode(title='Last name', validator=MinLengthValidator(5)

            # This syntax however is just like above. Be aware that your
            # attribute name will not be automatically added since your
            # explicitly defining args
            address = EntryNode(validator=
                        Check(MinLengthValidator(9), 'address'))

    If neither kwargs or args are specified and cannot be implicitly determined
    an exception will be thrown. """

    def __init__(self, validator, *args, **kwargs):
        self.validator = validator
        if not args:
            self.args = []
        else:
            self.args = list(args)

        if not kwargs:
            self.kwargs = {}
        else:
            self.kwargs = kwargs

        self._attr_name = None
        self.resolved = False

    def resolve_attr_names(self, data, form):
        """ Called internally by the validation methods this resolves all arg
        and kwarg strings to their respective `Node` objects and replaces them
        with a KeyedTuple containing the submitted data and the Node object
        reference.
        :param form: A reference to the Form class that the check is being
        resolved to.
        :param data: The full form data dictionary submitted for validation.
        """

        # Process args
        for key, arg in enumerate(self.args):
            # We need to get our node information.
            # If already resolved, just pull from arg
            if not self.resolved:
                self.args[key] = form.get_by_attr(arg)
            self.args[key].resolve_data(data)

        # Process kwargs
        for key, val in self.kwargs.iteritems():
            # We need to get our node information. If already resolved, just
            # pull from arg
            if not self.resolved:
                self.kwargs[key] = form.get_by_attr(val)
            self.kwargs[key].resolve_data(data)

        self.resolved = True

    def validate(self):
        """ Called by the validation routines. Allows the Check to specify
        parameters that will be passed to our Validation method.
        """
        return self.validator(*self.args, **self.kwargs)



from yota.exceptions import ValidatorNotCallableException, FormDataAccessException
from yota.renderers import JinjaRenderer
from yota.processors import FlaskPostProcessor
from yota.nodes import LeaderNode, Node
from yota.validators import Check
import json
import copy

class TrackingMeta(type):
    def __init__(mcs, name, bases, dict):
        """ Process all of the attributes in the `Form` (or subclass)
        declaration and place them accordingly. This builds the internal
        _node_list and _validation_list. """

        t = {}
        mcs._validation_list = []
        for name, value in dict.items():
            if isinstance(value, Node):
                value._attr_name = name
                t[value._create_counter] = value
                if hasattr(value, 'validators'):
                    if not isinstance(value.validator, tuple) and \
                       not isinstance(value.validator, list):
                        value.validators = [value.validators,]
                    for validator in value.validators:
                        # shorthand for adding a validation tuple
                        c = Check(validator, name)
                        mcs._validation_list.append(c)
            elif isinstance(value, Check):
                # if we've found a validation tuple
                value._attr_name = name
                mcs._validation_list.append(value)

        mcs._node_list = []
        for i, value in sorted(t.items()):
            # keeps track of the order of items for actual rendering
            mcs._node_list.append(value)

class Form(object):
    """ This is the base class that all user defined forms should inherit from,
    and as such it is the main way to access functionality in Yota. It
    provides the core functionality involved with setting up and
    rendering the form.

    :param context: This is a context specifically for the special form open and form close nodes, canonically called start and close.

    :param g_context: This is a global context that will be passed to all nodes in rendering thorugh their rendering context as 'g' variable.

    :param start_template: The template used when automatically injecting a start Node. See :attr:`yota.Form.auto_start_close` for more information.

    :param close_template: The template used when automatically injecting a close Node. See :attr:`yota.Form.auto_start_close` for more information.

    :param auto_start_close: Dictates whether or not start and close Nodes will be automatically appended/prepended to your form. Note that this must be set via __init__ or your class definition since it must be set before __init__ for the Form is run.

    :param hidden: A dictionary of hidden key/value pairs to be injected into the form.  This is frequently used to pass dynamic form parameters into the validator.

    """

    __metaclass__ = TrackingMeta
    _renderer = JinjaRenderer
    """ This is a class object that is used to perform the actual rendering
    steps, allowing different rendering engines to be swapped out. More about
    this in the section :class:`Renderer` """
    _processor = FlaskPostProcessor
    """ This is a class that performs post processing on whatever is passed in
    as data during validation. The intended purpose of this was to write
    processors that translated submitted form data from the format of the web
    framework being used to a format that Yota expects. It also allows things
    like filtering stripping characters or encoding all data that enters a
    validator. """


    def __new__(cls, **kwargs):
        """ We want our created Form to have a copy of the original
        form list so that dynamic additions to the list do not
        effect all Form instances """

        c = super(Form, cls).__new__(cls, **kwargs)
        c._node_list = copy.deepcopy(cls._node_list)
        for n in c._node_list:
            setattr(c, n._attr_name, n)
        c._validation_list = copy.deepcopy(cls._validation_list)
        for n in c._validation_list:
            if n._attr_name:
                setattr(c, n._attr_name, n)
        return c

    def __init__(self,
                 name=None,
                 auto_start_close=True,
                 start_template='form_open',
                 close_template='form_close',
                 g_context=None,
                 context=None,
                 start=None,
                 close=None,
                 form_class=None,
                 **kwargs):
        self.auto_start_close = auto_start_close
        self.start_template = start_template
        self.close_template = close_template
        self.g_context = g_context if g_context else {}
        self.context = context if context else {}

        # set a default for our name to the class name
        self.name = name if name else self.__class__.__name__

        # Set a default form class
        self.form_class = form_class if form_class else "form-horizontal"

        # passes everything to our rendering context and updates params
        self.context.update(kwargs)

        # since our default id is based off of the parent id
        # we can pass it in here
        for n in self._node_list:
            n.set_identifiers(self.name)

        # Add our open and close form to the end of the tmp lst
        if not start:
            self.insert(0, LeaderNode(template=self.start_template,
                         _attr_name='start', css_class=self.form_class,
                         **self.context))
        else:
            self.insert(0, start)

        if not close:
            self.insert(-1, LeaderNode(template=self.close_template,
                       _attr_name='close',
                       **self.context))
        else:
            self.insert(-1, close)



    def render(self):
        """ Runs the renderer to parse templates of nodes and generate the form
        HTML.

        :returns: A string containing the generated output.
        """

        return self._renderer().render(self._node_list, self.g_context)

    def insert(self, position, new_node_list):
        """ Inserts a :class:`Node` object or a list of objects at the specified
        position into the :attr:`Form._node_list` of the form. Index -1 is an
        alias for the end of the list.  After insertion the
        :meth:`Node.set_identifiers` will be called to generate identification
        for the :class:`Node`. For this to function, :attr:`Form._attr_name`
        must be specified for the node prior to insertion. """

        # check to allow passing in just a node
        if isinstance(new_node_list, Node):
            new_node_list = (new_node_list,)

        for i, new_node in enumerate(new_node_list):
            if position == -1:
                self._node_list.append(new_node)
            else:
                self._node_list.insert(position + i, new_node);
            setattr(self, new_node._attr_name, new_node)
            new_node.set_identifiers(self.name)

    def insert_after(self, prev_attr_name, new_node_list):
        """ Runs through the internal node structure attempting to find a
        :class:`Node` object whos :attr:`Node._attr_name` is prev_attr_name and
        inserts the passed node after it. If `prev_attr_name` cannot be matched it
        will be inserted at the end. Internally calls :meth:`Form.insert` and
        has the same requirements of the :class:`Node`.

        :param prev_attr_name: The attribute name of the `Node` that you would like to insert after.
        :type prev_attr_name: string
        :param new_node_list: The :class:`Node` or list of Nodes to be inserted.
        :type new_node_list: Node or list of Nodes """

        # check to allow passing in just a node
        if isinstance(new_node_list, Node):
            new_node_list = (new_node_list,)

        # Loop through our list of nodes to find where to insert
        for index, node in enumerate(self._node_list):
            # found!
            if node._attr_name == prev_attr_name:
                for i, new_node in enumerate(new_node_list):
                    self._node_list.insert(index + i + 1, new_node);
                    setattr(self, new_node._attr_name, new_node)
                    new_node.set_identifiers(self.name)
                break
        else:
            # failover append if not found
            for new_node in new_node_list:
                self._node_list.append(new_node)

    def get_by_attr(self, name):
        """ Safe accessor for looking up a node by :attr:`Node._attr_name` """
        try:
            attr = getattr(self, name)
        except:
            return None
        if isinstance(attr, Node):
            return attr
        return None

    def error_header_generate(self, errors, block):
        """ This method is called when any validators on the form fail to pass.
        The method should generate a dictionary that will be passed to your
        error renderer, whether that be javascript callbacks or re-rendering the
        form with error info in the rendering context.

        :param errors: This will be a list of all other Nodes that have errors.
        :param block: Whether or not the form submission will be blocked.
        :type block: boolean

        By default this function does nothing, but a common method of
        implementing it may be as follows:
        .. code-block:: python

            self.start.add_error({'message':
                                  'Please resolve the errors below to continue.'})

        This will provide a simple error message to your start Node.
        """
        pass


    def _gen_validate(self, data, piecewise=False):
        """ This is an internal utility function that does the grunt work of
        running validation logic for a :class:`Form`. It is called by the other
        primary validation methods. """
        for node in self._node_list:
            node.errors = []

        # Allows user to set a modular processor on incoming data
        data = self._processor().filter_post(data)

        # accumulate a set of nodes that may contain errors
        node_set = set()
        # loop over our checks and run our validators
        for check in self._validation_list:
            # try to iterate over their validators
            try:
                check.resolve_attr_names(data, self)
            except FormDataAccessException as e:
                # ignore the exception if we're in piecewise mode
                if not piecewise:
                    raise e
                else:
                    continue
            try:
                # Run our validator
                check.validate()
            except TypeError as e:
                raise ValidatorNotCallableException("Validators provided must "
                "be callable, type '{0}' instead. Caused by {1}". \
                        format(type(check.validator), e))
            # populate our set with potentially effected nodes
            node_set.update(check.kwargs.itervalues())
            node_set.update(check.args)

        block = False
        # a list to hold Nodes that actually have errors
        error_node_list = []
        if node_set:
            # Set the error value of the node to equal the dictionary that
            # is returned by the validator

            for node in node_set:
                # slightly confusing way of setting our block = True by
                # default
                if node.errors:
                    error_node_list.append(node)
                else:
                    continue

                for error in node.errors:
                    block |= error.get('block', True)

        return block, error_node_list


    def json_validate(self, data, piecewise=False):
        """ The same as :meth:`Form.validate_render` except the errors are loaded into
        a JSON string to be passed back as a query result. This output is
        designed to be used by the Yota Javascript library.

        :param piecewise: If set to True, the validator will silently ignore validator for which it has insufficient information. This is designed to be used for the AJAX piecewise validation function, although it does not have to be.
        :type piecewise: boolean
        """

        # Allows user to set a modular processor on incoming data
        data = self._processor().filter_post(data)

        errors = {}
        block, invalid = self._gen_validate(data, piecewise=piecewise)
        # loop over our nodes
        for node in invalid:
            errors[node.id] = node.errors

        # if needed we should run our all form message generator and return
        # json encoded error message
        retval = {'success': not block}
        if len(errors) > 0:
            errors['start'] = self.error_header_generate(errors, block)
        retval['errors'] = errors
        return json.dumps(retval)

    def validate_render(self, data):
        """ Runs all the validators on the `data` that is passed in and
        returns a re-render of the :class:`Form`. Since validators are designed
        to pass error information in through the :attr:`Node.errors` attribute
        then this error information is in turn availible through the rendering
        context.

        :param data: The data to be passed through the `Form._processor`. If the data is in the form of a dictionary where the key is the 'name' of the form field and the data is a string then no post-processing is neccessary.
        :type data: dictionary
        """

        # Allows user to set a modular processor on incoming data
        data = self._processor().filter_post(data)

        block, invalid = self._gen_validate(data)

        # run our form validators at the end
        if len(invalid) > 0:
            self.error_header_generate(invalid, block)

        return self.render()




import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

requires = [
]

setup(name='yota',
      version='0.1.3',
      description='A form library with a focus on simplicity',
      long_description="",
      author='',
      url='',
      packages=find_packages('src'),
      include_package_data=True,
      install_requires=requires,
      package_dir = {'': 'src'},
      )
